#!/usr/bin/perl

# this script is used by the perl6 infrastructure to manage precompiled files of
# perl6 module packages. after the installation of a package, this is run
# from the module package postinst script and installs the perl6 sources from the 
# module package into the perl6 code repository, creating precomp files in the
# process. All files created are tracked so that they can be cleaned up on
# package removal. A rakudo upgrade may need to redo the precompilation when the
# rakudo package is no longer compatible with the existing precomp files, this
# script also takes care of that.

use strict;
use warnings;
use 5.20.1;

use Getopt::Long;
use Path::Tiny;
use Graph;

# this is determined and injected at package build times, and represents 
# the compiler id of this rakudo version. precomp files are only 
# compatible between rakudos that have the same compiler id.
my $compiler_id = "4148F3E09E51EB6F719229EDF5BC67B567E37316";

my $varlib = path('/var/lib/perl6/modules');
my $vendor = path('/usr/lib/perl6/vendor');
my $p6src  = path('/usr/share/perl6/debian-sources');

sub debug_say($) {
    if (defined $ENV{'RAKUDO_HELPER_DEBUG'}) {
        say @_;
    }
}

sub install($) {
    my ($module) = @_;
    debug_say("installing perl6 module $module...");
    my $tmpdir = Path::Tiny->tempdir;
    my $build_dir = $tmpdir->child('build');
    $build_dir->mkpath;
    my $src_dir = $p6src->child($module);
    $ENV{'RAKUDO_RERESOLVE_DEPENDENCIES'} = 0;
    my @cmd = qq!perl6 /usr/share/perl6/tools/install-dist.p6 --from=$src_dir --to=$build_dir --for=vendor!;
    system(@cmd) == 0 
        or die "module $module compilation failed: $?";

    @cmd = qq!find $build_dir -type d -empty -delete!;
    system(@cmd) == 0 
        or die "could not clean up empty dirs after module compilation for $module: $?";
    my @resulting_files;
    $build_dir->visit(sub {
        my ($path, $state) = @_;
        return if $path->is_dir;
            push(@resulting_files, $path =~ s/.*\/build\///r);
        }, { recurse => 1 });
    my $fh;
    open($fh, "> $varlib/$module.list")
        or die "could not open file in module compilation for $module: $!";
    print $fh join("\n", @resulting_files);
    close($fh);
    @cmd = qq!cp -r $build_dir/* $vendor!;
    system(@cmd) == 0 
        or die "could not copy files in module compilation for $module: $?";
}

sub remove($) {
    my ($module) = @_;
    if (-e "$varlib/$module.list") { 
        debug_say("removing perl6 module $module...");
        my $fh;
        open($fh, "< $varlib/$module.list")
            or die "could not open file in removal of module $module: $!";
        while (<$fh>) {
            chomp;
            # XXX this feels a bit unsafe...
            unlink "$vendor/$_";
        }
        close($fh);
        unlink "$varlib/$module.list";
        my @cmd = qq!find $vendor/short $vendor/precomp $vendor/dist $vendor/sources -type d -empty -delete!;
        system(@cmd) == 0 
            or die "could not remove empty directories in removal of module $module: $?";
    }
}

sub reinstall_all() {
    # remove all existing precompiled files
    debug_say("reinstalling all perl6 modules on system...");
    foreach my $list ($varlib->children(qr/\.list$/)) {
        my $module = $list->basename =~ s/\.list$//r;
        debug_say("deleting old precompiled files of module $module");
        remove($module);
    }
    # now build a graph of module dependencies
    my $g = Graph->new;
    foreach my $cdir ($p6src->children) {
        next unless $cdir->is_dir;
        my $module = $cdir->basename;
        foreach my $dep (path("$p6src/$module/$module.p6deps")->lines({ chomp => 1 })) {
            debug_say("  $module -> $dep");
            $g->add_edge($dep => $module);
        }
    }
    foreach my $module ($g->topological_sort) {
	if ($module) {
	    if ($p6src->child($module)->is_dir) {
                install($module);
	    }
	    else {
	        debug_say("warning: $module shows up in package dependency graph but is not installed");
	    }
        }
    }
}

sub usage {
    print <<'EOF';
Rakudo/Perl 6 Helper script
Usage: rakudo-helper.pl [arguments]
Where [arguments] can be one of:
  --install -i <package> install perl modules from specified package
                         into rakudo distribution
  --remove -r <package>  remove files created by --install
  --reinstall-all        reinstall all perl6 module packages on the system
  --compiler-id -c       print the rakudo compiler id
  --help -h              show this text
EOF
}

my $arg_install = undef;
my $arg_remove = undef;
my $arg_reinstall_all;
my $arg_compiler_id;
my $arg_help;

GetOptions(
        "--install=s" => \$arg_install,
        "--remove=s" => \$arg_remove,
        "--reinstall-all|R" => \$arg_reinstall_all,
        "--compiler-id" => \$arg_compiler_id,
        "--help" => \$arg_help,
    )
    or do {
        print STDERR "Could not parse cmdline arguments";
        usage();
        exit 1;
    };

if ($arg_help) {
    usage();
    exit 0;
}
elsif ($arg_compiler_id) {
    say $compiler_id;
}
elsif (defined $arg_install) {
    install($arg_install);
}
elsif (defined $arg_remove) {
    remove($arg_remove);
}
elsif (defined $arg_reinstall_all) {
    reinstall_all();
}
else {
    print STDERR "No argument specified";
    usage();
    exit 1;
}
